/* UADE - Unix Amiga Delitracker Emulator
 * Copyright 2000-2002, Heikki Orsila <heikki.orsila@ee.tut.fi>
 * See http://ee.tut.fi/~heikki/uade.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "events.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "debug.h"

#include "uade.h"
#include "uade-os-unix.h"

#include "decrunch/decrunch.h"
#include "../plugindir/amifilemagic.h"

#include <netinet/in.h>

#include "../osdep/strlrep.c"

static int uade_calc_reloc_size(unsigned int *src);
static int uade_safe_load(int dst, FILE *file, int maxlen);
static void uade_put_long(int addr,int val);
static void uade_siginthandler(int);
static void uade_signalhandler(int);
static void uade_getmessage(void);
static void uade_waitformessage(int msg);
static void uade_set_subsong(int);
static void uade_set_ntsc(int);
static void uade_set_songend(int songendpossible);
static void uade_send_amiga_message(int);

static void uade_enque_song(const char *filename);
static void check_name_extension(char *filename, char *playername);

static void uade_print_help(int problemcode);

static const int SCORE_MODULE_ADDR   = 0x100;
static const int SCORE_MODULE_LEN    = 0x104;
static const int SCORE_PLAYER_ADDR   = 0x108;
static const int SCORE_RELOC_ADDR    = 0x10C;
static const int SCORE_USER_STACK    = 0x110;
static const int SCORE_SUPER_STACK   = 0x114;
static const int SCORE_FORCE         = 0x118;
static const int SCORE_SET_SUBSONG   = 0x11c;
static const int SCORE_SUBSONG       = 0x120;
static const int SCORE_NTSC          = 0x124;
static const int SCORE_MODULE_NAME_ADDR = 0x128;
static const int SCORE_HAVE_SONGEND  = 0x12C;
static const int SCORE_POSTPAUSE     = 0x180;
static const int SCORE_PREPAUSE      = 0x184;
static const int SCORE_DELIMON       = 0x188;
static const int SCORE_EXEC_DEBUG    = 0x18C;
static const int SCORE_VOLUME_TEST   = 0x190;
static const int SCORE_DMA_WAIT      = 0x194;

static const int SCORE_INPUT_MSG     = 0x200;
static const int SCORE_MIN_SUBSONG   = 0x204;
static const int SCORE_MAX_SUBSONG   = 0x208;
static const int SCORE_CUR_SUBSONG   = 0x20C;

static const int SCORE_OUTPUT_MSG    = 0x300;

static int uade_highmem = 0x200000;

/* is uade commanded by external application? */
int uade_slavemode = 0;
uade_msgstruct *uade_struct;

char uade_playername[256];       // filename of eagleplayer
char uade_modulename[256];       // filename of song
char uade_scorename[256];        // filename of score file
char uade_mapname[256];          // filename of sharemem file

char score_playername[256];      // playername that eagleplayer reports
char score_modulename[256];      // modulename that eagleplayer reports
char score_formatname[256];	 // formatname that eagleplayer reports

static int uade_setsubsongboolean = 0;
static int uade_subsong = 0;
static int uade_execdebugboolean = 0;
static int uade_force_by_default = 0;
static int uade_use_ntsc = 0;
static int uade_songendpossible = 1;
static int uade_repeat = 0;

static int uade_max_subsong;
static int uade_min_subsong;
static int uade_cur_subsong;
static int uade_one_subsong = 0;

/* don't touch */
static int uade_dmawait = 0;
static int voltestboolean = 0;

static int playerforced = 0;

/* output sound data to pipe */
int uade_usingoutpipe = 0;
/* This should be non-zero if native sound output shouldn't be used.
   This is set to non-zero in slavemode and outpipemode              */
int uade_no_native_sound = 0;

int uade_reboot;

int uade_messages;       // number of messages in msgtable
static int msgtable[20]; // contains signals sent from master

static struct sigaction sigact;

int uade_mixpar = 0;

struct sn_t {
  struct sn_t *next;
  char filename[256];
};
typedef struct sn_t songnode_t;
songnode_t *SongList = 0;

/* contains uade's command line name */
static char *uadecmdlinename = 0;


void uade_option(int argc, char** argv) {
  int i;

  uadecmdlinename = strdup(argv[0]);

  uade_mapname[0] = 0;
  uade_playername[0] = 0;

  uade_get_path(uade_scorename, UADE_PATH_SCORE, sizeof(uade_scorename));

  for (i = 1; i < argc;) {
    // check for playername parameter
    if (strcasecmp(argv[i], "-P") == 0) {
      if((i+1) < argc) {
	strcpy(uade_playername, argv[i+1]);
	playerforced = 1;
	fprintf(stderr,"-P parameter: %s\n", uade_playername);
	i+=2;
      } else {
	fprintf(stderr, "Player name missing for -P parameter\n\n");
	uade_print_help(1);
	exit(-1);
      }

      // check for modulename parameter
    } else if (strcasecmp(argv[i], "-M") == 0) {
      if((i+1) < argc) {
	fprintf(stderr, "-M parameter: %s\n", argv[i+1]);
	uade_enque_song(argv[i+1]);
	i+=2;
      } else {
	fprintf(stderr, "Song name missing for -M parameter\n\n");
	uade_print_help(1);
	exit(-1);
      }

      // check for scorename parameter
    } else if(strcasecmp(argv[i], "-S") == 0) {
      strcpy(uade_scorename, argv[i+1]);
      fprintf(stderr, "-S parameter: %s\n", uade_scorename);
      i+=2;

      // check if this prog is requested for slave
      // that let's external programs use this program as
      // a player
    } else if (strcmp(argv[i], "-slave") == 0) {
      strcpy(uade_mapname, argv[i+1]);
      uade_slavemode = 1;
      uade_no_native_sound = 1;
      i+=2;

    } else if (strcmp(argv[i], "-outpipe") == 0) {
      fprintf(stderr, "-outpipe parameter: %s\n", argv[i+1]);
      uade_usingoutpipe = 1;
      uade_no_native_sound = 1;
      uade_init_outpipe(argv[i+1]);
      i+=2;

    } else if (strcmp(argv[i], "-force") == 0) {
      fprintf(stderr, "-force parameter\n");
      uade_force_by_default = 1;
      i++;

    } else if (strcmp(argv[i], "-execdebug") == 0) {
      fprintf(stderr, "-execdebug parameter\n");
      uade_execdebugboolean = 1;
      i++;

    } else if (strcmp(argv[i], "-voltest") == 0) {
      fprintf(stderr, "-voltest parameter\n");
      voltestboolean = 0x12345678;
      i++;

    } else if (strcmp(argv[i], "-dmawait") == 0) {
      uade_dmawait = atoi(argv[i+1]);
      fprintf(stderr, "setting dmawait = %d rasterlines\n", uade_dmawait);
      uade_dmawait += 0x12340000;
      i += 2;

    } else if (strcmp(argv[i], "-ntsc") == 0) {
      fprintf(stderr, "-ntsc parameter (default is PAL)\n");
      uade_use_ntsc = 1;
      i++;

    } else if (strcmp(argv[i], "-noend") == 0) {
      fprintf(stderr, "-noend parameter (player can't force song to end\n");
      uade_songendpossible = 0;
      i++;

    } else if (strcmp(argv[i], "-sub") == 0) {
      uade_subsong = atoi(argv[i+1]);
      if((*argv[i+1] != '+') && (*argv[i+1] != '-')) { 
	uade_setsubsongboolean = 1;
	fprintf(stderr, "-sub parameter parsed, setting subsong %d\n", uade_subsong);
      } else {
	uade_setsubsongboolean = 2;
	fprintf(stderr, "-sub parameter parsed, setting subsong %d+min_subsong\n", uade_subsong);
      }
      i+=2;

    } else if (strcmp(argv[i], "-one") == 0) {
      fprintf(stderr, "-one parsed: playing at most one subsong / file\n");
      uade_one_subsong = 1;
      i++;

      /* -int is obsolete */
    } else if (strcmp(argv[i], "-int") == 0) {
      fprintf(stderr, "-int is an obsolete parameter!\n\n");
      i++;

    } else if (strcmp(argv[i], "--verbose") == 0) {
      i++;

    } else if (strcmp(argv[i], "--version") == 0) {
      fprintf(stdout,"UADE %s\n", uade_version());
      exit(0);

    } else if (strcmp(argv[i], "-varpid") == 0) {
      /* ADDED BY XIGH */
      uade_create_var_pid();
      i++;

    } else if(strcmp(argv[i], "-pan")==0) {
      fprintf(stderr, "-pan parameter parsed (%s)\n", argv[i+1]);
      uade_mixpar = (int) (256.0f/2.0f*atof(argv[i+1]));
      i+=2;

    } else if(strcmp(argv[i], "-config")==0) {
      i++;

    } else if(strcmp(argv[i], "-repeat")==0) {
      uade_repeat = 1;
      i++;

    } else if(strcmp(argv[i], "--help")==0 || strcmp(argv[i], "-h")==0) {
      uade_print_help(0);
      exit(0);

    } else {

      uade_enque_song(argv[i]);

      i++;
    }
  }

  if(!uade_slavemode) {
    // if songlist is empty
    if(!SongList) {
      if(uade_playername[0] != 0) {
	/* queue zero string as module, only player is used */
	uade_enque_song(strdup(""));
      } else {
	uade_print_help(2);
	exit(-1);
      }
    }
    /* create CTRL-C break for cmdline tool */
    uade_create_signalhandler(uade_siginthandler, 2);
  }

  if(uade_slavemode) {

    uade_struct = (uade_msgstruct*) uade_mmap_file(uade_mapname);
    if(!uade_struct) {
      fprintf(stderr, "uade.c/uade: couldn't mmap file\n");
      exit(EXIT_FAILURE);
    }

    uade_struct->minsubsong = 0;
    uade_struct->maxsubsong = 0;
    uade_reboot = 1;

    /* clear messages in msgtable */
    uade_messages = 0;
    for(i=0; i<20; i++) msgtable[i] = 0;

    /* setup a signal handler for communication with xmms plugin */
    uade_create_signalhandler(uade_signalhandler, 1);

    /* leave a watchdog that checks if XMMS plugin dies (polls plugin
       task with a kill(pluginpid,0) signal every 0.5 seconds). If XMMS
       plugin dies, watchdog kills UAE and itself */
    uade_create_watchdog(uade_struct, uade_mapname);
  }
}


static void uade_print_help(int problemcode) {
  switch(problemcode) {
  case 0:
    /* just for printing help */
    break;
  case 1:
    fprintf(stderr, "Illegal parameters for UADE\n\n");
    break;
  case 2:
    fprintf(stderr, "No songs given as parameters for UADE\n\n");
    break;
  default:
    fprintf(stderr, "Unkown error\n");
    break;
  }
  fprintf(stderr, "UADE usage:\n");
  fprintf(stderr, " %s [OPTIONS] [FILE1] [FILE2] ...\n\n", uadecmdlinename);

  fprintf(stderr, "options:\n");
  fprintf(stderr, " -M songname\tSets songname to be played\n");
  fprintf(stderr, " -P playername\tSets eagleplayer to be used (you can use this for custom songs)\n");
  fprintf(stderr, " -sub number\tSets first subsong number to be played\n");
  fprintf(stderr, " -force\t\tForces eagleplayer to play a given song if it's not recognized\n");
  fprintf(stderr, " -repeat\tLoop song list to be played from beginning after the end\n");
  fprintf(stderr, " -pan value\tSets panning value in the range of [0.00, 1.00] (mix channels)\n");
  fprintf(stderr, " -one\t\tPlay at most one subsong / file\n");
  fprintf(stderr, " -noend\t\tThe first song will be played indefinetely (probably)\n");
  fprintf(stderr, " -ntsc\t\tForce Amiga into NTSC mode (may affect timing)\n");
  fprintf(stderr, " -outpipe x\tDump all sound output to file descriptor x (x=1 => stdout)\n");
  fprintf(stderr, " -S scorename\tForces scorename file to be used as 'sound core' (experts only)\n");
}


static void uade_enque_song(const char *fname) {
  songnode_t *s = SongList;
  songnode_t *new;
  if(uade_if_normal_file(fname) || strlen(fname) == 0) {
    new = malloc(sizeof(songnode_t));
    new->next = 0;
    strcpy(new->filename, fname);
    if(s) {
      while(s->next) s = s->next;
      s->next = new;
    } else {
      SongList = new;
    }
  } else {
    fprintf(stderr,"uade: '%s' is not regular file\n", fname);
  }
}


static void check_name_extension(char *filename, char *playername) {
  FILE *formatsfile;
  FILE *songfile;
  int filesize = 5122;
  unsigned char buf[5122] = "";
  int status;
  char formatsfilename[1024];
  char playerdir[1024];
  int x;
  char a[1024], b[1024], pre[1024], post[1024];
  char *c, *filename2;

  *playername = 0;

  if(!uade_get_path(formatsfilename, UADE_PATH_FORMATSFILE, sizeof(formatsfilename)))
    return;
  formatsfile = fopen(formatsfilename, "rb");

  /* get player directory */
  uade_get_path(playerdir, UADE_PATH_PLAYERDIR, sizeof(playerdir));

  pre[0]='\0';
  post[0]='\0';

  filename2 = uade_get_basename(filename);
  c = strchr(filename2, (int) '.');
  if(c == 0) {
    goto closeffile;
  }
  strncpy(pre, filename2, c-filename2);
  pre[c-filename2] = '\0';              // make so that pre is null-terminated
  c = strrchr(filename2, (int) '.');
  strcpy(post, c+1);                    // post is null-terminated

  songfile = fopen (filename, "rb");
  if (!songfile) {
    fprintf (stderr, "can not open file %s\n", filename);
    return;
  }
  if ((status = decrunch ( &songfile,filename ) < 0 )) {
    fprintf (stderr, "decrunching error...\n");
    fclose (songfile);
    return;
  }
  fread  (buf, 1, 5122, songfile);
  fclose (songfile);
  filemagic (buf, pre, post);

  for(;;) {
    x = fscanf(formatsfile,"%s",a);
    if(x==0 || x==EOF) break;

    // check if this is a comment line (this skips comment lines)
    if(a[0]=='#') {
      while(a[0]!='\n') {
	x=fscanf(formatsfile,"%c",&a[0]);
	if(x==0 || x==EOF) break;
      }
      continue;
    }

    if(strcasecmp(a,"formats")==0) {
      for(;;) {
	x=fscanf(formatsfile,"%s",a);
	if(x==0 || x==EOF) break;
	if(strcasecmp("endformats",a)==0) break;
	
	// check if this is a comment line (this skips comment lines)
	if(a[0]=='#') {
	  while(a[0]!='\n') {
	    x=fscanf(formatsfile,"%c",&a[0]);
	    if(x==0 || x==EOF) break;
	  }
	  continue;
	}
	
	x=fscanf(formatsfile,"%s",b);
	if(x==0 || x==EOF) break;

	if(strcasecmp(pre,a)==0 || strcasecmp(post,a)==0) {
	  if(strcmp(b,"custom")==0) {
	    strcpy(playername, filename);
	    filename[0] = 0;
	  } else {
	    /* Find out player directory path */
	    strcpy(playername, playerdir);
	    strcat(playername, b);
	  }
	  goto closeffile;
	}
      }
      break;
    }
  }
 closeffile:
  fclose(formatsfile);
}


/* catch CTRL-C and cause REBOOT */
UADE_SIGNALHANDLER(uade_siginthandler) {
  int ltime;
  signum = 0;
  ltime = uade_timer_poll();
  if(uade_reboot || (ltime>0 && ltime<100)) {
    fprintf(stderr,"uade: exit forced with ctrl-c\n");
    exit(0);
  }
  uade_reboot = 1;
}


UADE_SIGNALHANDLER(uade_signalhandler) {
  int msg,msgs;
  char a = 0;
  int i;

  msg = uade_struct->touaemsgtype;
  uade_struct->touaemsgtype = 0;

  switch(msg) {
  case UADE_REBOOT:
    if(uade_reboot == 1) return;
    uade_reboot = 1;
    uade_clear_soundbuffer();
    break;
  case UADE_SETSUBSONG:
    uade_set_subsong(uade_struct->subsong);
    uade_struct->setsubsongboolean = 0;
    return;

  case UADE_PLAYERNAME:
  case UADE_MODULENAME:
  case UADE_SCORENAME:
    break;
  case UADE_NTSC:
    uade_set_ntsc(uade_struct->ntscbit);
    uade_send_amiga_message(UADE_NTSC);
    return;
  case UADE_SONGEND:
    uade_songendpossible = uade_struct->songendpossible;
    uade_set_songend(uade_songendpossible);
    return;
  default:
    fprintf(stderr,"uade_getmessage: Msg was not recognized (%d)\n",msg);
    break;
  }

  msgtable[msg]=1;

  /* count messages in msgtable */
  msgs=0;
  for(i=0;i<20;i++) msgs+=(msgtable[i]==0)?0:1;
  uade_messages=msgs;
  /* fprintf(stderr,"uae: getmessage: uade_messages=%d\n",uade_struct->messages); */
}

void uade_waitformessage(int msg) {
  int sl = 0;
  // fprintf(stderr,"uade: uade_waitformessage: %d\n",msg);
  while(1) {
    if(msgtable[msg] != 0) {
      uade_handle_message(msg);
      break;
    }
    uade_usleep(1000);
    sl += 1000;
  }
  fprintf(stderr,"uade: uade_waitformessage: slept %d\n",sl/1000);
}


void uade_handle_message(int msg) {
  int i,msgs;
  msgtable[msg] = 0;
  msgs = 0;
  for(i=0;i<20;i++)
    msgs += (msgtable[i]==0) ? 0:1;
  uade_messages=msgs;
}


void uade_prerun(void) {
  /* don't load anything under 0x1000 (exec at $1000) */
  int modnameaddr = 0x00400;
  int scoreaddr   = 0x01000;
  int userstack   = 0x08500;
  int superstack  = 0x08f00;
  int playeraddr  = 0x09000;
  int relocaddr   = 0x40000;   /* default, recalculate this */
  int modaddr     = 0x80000;   /* default, recalculate this */
  int len;
  FILE* file;
  uae_u8* memp;
  int bytesread;
  /* decrunch variables */
  FILE* file_cr;
  char *modfile;
  int status;

  /* IMPORTANT:
     It seems that certain players don't work totally reliably if memory
     contains trash from previous songs. To be certain that each song is
     played from the same initial state of emulator we clear the memory
     from 0x400 to 0x100000 each time a new song is played */

  uade_highmem = 0x0;
  while(uade_highmem < 0x200000) {
    if(!valid_address(0, uade_highmem + 0x10000))
      break;
    uade_highmem += 0x10000;
  }
  if(uade_highmem < 0x200000)
    fprintf(stderr, "uade: warning: highmem < 0x200000!\n");

  if(valid_address(0x400, 0x100000-0x400)) {
    memset(get_real_address(0x400), 0, 0x100000-0x400);
  } else {
    fprintf(stderr, "uade: prerun: invalid memory range for clearing\n");
  }

  if(uade_slavemode) {
    /* RACE CONDITION BEGINS */
    /* uade executable is ready to receive commands */
    uade_struct->uade_inited_boolean = 1;
    /* sleep() is interrupted by a SIGHUP from XMMS plugin */
    while(uade_struct->loadnewsongboolean == 0) uade_sleep(2);
    /* RACE CONDITION ENDS */

    strcpy(uade_playername, uade_struct->playername);
    strcpy(uade_modulename, uade_struct->modulename);
    strcpy(uade_scorename, uade_struct->scorename);

    uade_struct->loadnewsongboolean = 0;

    uade_force_by_default = uade_struct->force_by_default;
    if(uade_struct->setsubsongboolean) {
      uade_struct->setsubsongboolean = 0;
      uade_setsubsongboolean = 1;
      uade_subsong = uade_struct->subsong;
    }
    uade_struct->minsubsong = uade_struct->maxsubsong = 0;
    uade_songendpossible = uade_struct->songendpossible;

  } else {

  takenextsong:

    /* if no more songs in the play queue, quit cmdline tool */
    if(!SongList) {
      fprintf(stderr,"uade: no more songs to play\n");
      exit(EXIT_SUCCESS);

    } else {
      char playername[256];
      char tempname[256];
      strcpy(uade_modulename, SongList->filename);
      strcpy(tempname, SongList->filename);

      /* uade can be forced to use 'uade_playername' initialized with
	 -P switch for all played songs. */
      if(!playerforced) {
	check_name_extension(uade_modulename, playername);
	/* playername is empty string if no playername was found */
	if(playername[0]!=0) {
	  strcpy(uade_playername, playername);
	} else {
	  fprintf(stderr,"uade: file %s: unknown format\n", SongList->filename);
	}
      }
      SongList = SongList->next;

      if(uade_repeat) {
	uade_enque_song(tempname);
      }
    }

    if(uade_playername[0] == 0)
      goto takenextsong;

    uade_cur_subsong = uade_min_subsong = uade_max_subsong = 0;
  }

  uade_set_songend(uade_songendpossible);

  if(uade_execdebugboolean)
    uade_put_long(SCORE_EXEC_DEBUG, 0x12345678);
  else
    uade_put_long(SCORE_EXEC_DEBUG, 0);

  uade_put_long(SCORE_VOLUME_TEST, voltestboolean);
  uade_put_long(SCORE_DMA_WAIT, uade_dmawait);

  /* open player */
  file = fopen(uade_playername, "rb");
  if(file) {
    if ((status = decrunch (&file, uade_playername)) < 0) {
      fprintf (stderr, "decrunching error...");
      fclose(file); file = 0;
      goto setupfailure;
    } else {
      bytesread = uade_safe_load(playeraddr, file, uade_highmem - playeraddr);
      fclose(file); file = 0;
      fprintf(stderr, "uade: player '%s' (%d bytes)\n", uade_playername, bytesread);
      /* set player executable address for relocator */
      uade_put_long(SCORE_PLAYER_ADDR, playeraddr);
      memp = get_real_address(playeraddr);
      len = uade_calc_reloc_size((unsigned int *) memp);
      if(len) {
	relocaddr  = ((playeraddr + bytesread) & 0x7FFFF000) + 0x2000;
	modaddr = ((relocaddr + len) & 0x7FFFF000) + 0x2000;
      } else {
	fprintf(stderr, "uade: problem with reloc calculation\n");
	fprintf(stderr, "uade: setting default addresses\n");
      }
      uade_put_long(SCORE_RELOC_ADDR, relocaddr);
    }
  } else {
    fprintf(stderr,"uade: couldn't load player %s\n", uade_playername);
  setupfailure:
    if(uade_slavemode) {
      uade_reboot = 0;
      uade_struct->songendboolean = 1;
      return;
    }
    exit(EXIT_FAILURE);
  }

  /* load module (modules are not always loaded) */
  uade_put_long(SCORE_MODULE_ADDR, modaddr);   /* set modules address */
  uade_put_long(SCORE_MODULE_LEN, 0);          /* set module size to zero */
  uade_put_long(SCORE_MODULE_NAME_ADDR, 0);    /* mod name address pointer */
  if(uade_modulename[0]) {
    if((file = fopen(uade_modulename, "rb"))) {
      if ((status = decrunch (&file, uade_modulename)) < 0) {
	fprintf (stderr, "uade: decrunching error...");
	fclose(file); file = 0;
	goto setupfailure;
      } else {
	bytesread = uade_safe_load(modaddr, file, uade_highmem - modaddr);
	fclose(file); file = 0;
	fprintf(stderr, "uade: module '%s' (%d bytes)\n", uade_modulename, bytesread);
	uade_put_long(SCORE_MODULE_LEN, bytesread);
	strlcpy(get_real_address(modnameaddr), uade_modulename, 256);
	uade_put_long(SCORE_MODULE_NAME_ADDR, modnameaddr);
      }
    } else {
      goto setupfailure;
    }
  } else {
    memp = get_real_address(modnameaddr);
    strlcpy(memp, uade_playername, 256);
    uade_put_long(SCORE_MODULE_NAME_ADDR, modnameaddr);
  }

  /* load sound core (score) */
  if((file = fopen(uade_scorename, "rb"))) {
    bytesread = uade_safe_load(scoreaddr, file, uade_highmem - scoreaddr);
    fclose(file); file = 0;
    /* fprintf(stderr, "uade: score  '%s' (%d bytes)\n", uade_scorename, bytesread); */
    m68k_setpc(scoreaddr);
    /* override bit for sound format checking */
    uade_put_long(SCORE_FORCE, uade_force_by_default);
    /* setsubsong */
    uade_put_long(SCORE_SET_SUBSONG, uade_setsubsongboolean);
    uade_put_long(SCORE_SUBSONG, uade_subsong);
    /* set ntscbit correctly */
    if(uade_slavemode) {
      uade_set_ntsc(uade_struct->ntscbit);
    } else {
      uade_set_ntsc(uade_use_ntsc);
    }
    /* pause bits (don't care!), for debugging purposes only */
    uade_put_long(SCORE_PREPAUSE, 0);
    uade_put_long(SCORE_POSTPAUSE, 0);
    /* set user and supervisor stack pointers */
    uade_put_long(SCORE_USER_STACK, userstack);
    uade_put_long(SCORE_SUPER_STACK, superstack);
    /* no message for score */
    uade_put_long(SCORE_OUTPUT_MSG, 0);
    if( (userstack - (scoreaddr + bytesread)) < 0x1000) {
      fprintf(stderr, "uade: stack over run warning!\n");
    }
  } else {
    goto setupfailure;
  }

  uade_setsubsongboolean = 0;

  if(!playerforced)
    uade_playername[0] = 0;

  uade_modulename[0] = 0;
  score_playername[0] = 0;
  score_modulename[0] = 0;
  if(uade_slavemode) {
    uade_struct->score_playername[0] = 0;
    uade_struct->score_playerauthor[0] = 0;
    uade_struct->score_modulename[0] = 0;
    uade_struct->score_formatname[0] = 0;
  }

  /* don't reboot after this routine has run unless SIGNAL has been
     caught... */
  uade_reboot = 0;
  if(uade_slavemode) {
    if(msgtable[UADE_REBOOT]) {
      int l;
      for(l=0;l<20;l++) {
	fprintf(stderr,"uade: TRIED TO REBOOT 2 TIMES !!!\n");
	fprintf(stderr," If you can make this bug happen again please submit");
	fprintf(stderr," a bug report.\n");
      }
    }
  }
}

static void uade_put_long(int addr, int val) {
  uae_u8 *p;
  p = get_real_address(addr+0);
  *p = (uae_u8) (val>>24 & 0xff);
  p = get_real_address(addr+1);
  *p = (uae_u8) (val>>16 & 0xff);
  p = get_real_address(addr+2);
  *p = (uae_u8) (val>>8 & 0xff);
  p = get_real_address(addr+3);
  *p = (uae_u8) (val & 0xff);
}

int uade_get_long(int addr) {
  uae_u8* ptr;
  int x = 0;
  ptr = get_real_address(addr);
  x += (*ptr)<<24;
  ptr = get_real_address(addr+1);
  x += (*ptr)<<16;
  ptr = get_real_address(addr+2);
  x += (*ptr)<<8;
  ptr = get_real_address(addr+3);
  x += (int) (*ptr);
  return x;
}

static int uade_safe_load(int dst, FILE *file, int maxlen) {
  const int bufsize = 4096;
  char buf[bufsize];
  int nbytes, len, off;
  len = bufsize;
  off = 0;
  if(maxlen < 0)
    return 0;
  while(!feof(file)) {
    if(maxlen < bufsize)
      len = maxlen;
    if(!(nbytes = fread(buf, 1, len, file)))
      break;
    if(!valid_address(dst + off, nbytes)) {
      fprintf(stderr, "uade: illegal load range [%x,%x)\n", dst + off, dst + off + nbytes);
      break;
    }
    memcpy(get_real_address(dst + off), buf, nbytes);
    off += nbytes;
    maxlen -= nbytes;
  }
  return off;
}

static void uade_safe_get_string(char *dst, int src, int maxlen) {
  int i = 0;
  while(1) {
    if(i >= maxlen) {
      fprintf(stderr, "uade: safe_get_string: i over maxlen\n");
      break;
    }
    if(!valid_address(src + i, 1)) {
      fprintf(stderr, "uade: safe_get_string: invalid memory range\n");
      break;
    }
    dst[i] = *((char *) get_real_address(src + i));
    i++;
  }
  if(maxlen > 0) {
    if(i < maxlen) {
      dst[i] = 0;
    } else { 
      fprintf(stderr, "uade: safe_get_string: null termination warning!\n");
      dst[maxlen-1] = 0;
    }
  }
}

void uade_get_amiga_message(void) {
  uae_u8 *ptr;
  uae_u8 *nameptr;
  FILE *file;
  int x;
  int mins, maxs, curs;
  int status;
  int src, dst, off, len;

  /* get input message type */
  x = uade_get_long(SCORE_INPUT_MSG);

  switch(x) {
  case UADE_SONGEND:
    fprintf(stderr,"uade: song end (player)\n");
    if(uade_slavemode) {
      uade_struct->songendboolean = 1;
    } else {
      if(uade_max_subsong && !uade_one_subsong) {
	if(uade_cur_subsong < uade_max_subsong) {
	  uade_cur_subsong++;
	  uade_set_subsong(uade_cur_subsong);
	  break;
	}
      }
      uade_reboot = 1;
    }
    break;
  case UADE_SUBSINFO:
    mins = uade_get_long(SCORE_MIN_SUBSONG);
    maxs = uade_get_long(SCORE_MAX_SUBSONG);
    curs = uade_get_long(SCORE_CUR_SUBSONG);
    fprintf(stderr, "uade: subsong info: min subsong: %d max subsong: %d cur subsong: %d\n", mins, maxs, curs);
    if(uade_slavemode) {
      uade_struct->minsubsong = mins;
      uade_struct->maxsubsong = maxs;
      uade_struct->subsong    = curs;
    } else {
      uade_min_subsong = mins;
      uade_max_subsong = maxs;
      uade_cur_subsong = curs;
    }
    break;
  case UADE_PLAYERNAME:
    strlcpy(score_playername, get_real_address(0x204), 25);
    fprintf(stderr,"uade: playername: %s\n", score_playername);
    if(uade_slavemode)
      strlcpy(uade_struct->score_playername, score_playername, 25);
    break;
  case UADE_MODULENAME:
    strlcpy(score_modulename, get_real_address(0x204), sizeof(score_modulename));
    fprintf(stderr,"uade: modulename: %s\n", score_modulename);
    if(uade_slavemode)
      strlcpy(uade_struct->score_modulename, score_modulename, sizeof(uade_struct->score_modulename));
    break;
  case UADE_FORMATNAME:
    strlcpy(score_formatname, get_real_address(0x204), sizeof(score_formatname));
    fprintf(stderr,"uade: formatname: %s\n", score_formatname);
    if(uade_slavemode)
      strlcpy(uade_struct->score_formatname, score_formatname, sizeof(uade_struct->score_formatname));
    break;
  case UADE_GENERALMSG:
    fprintf(stderr,"uade: general message: %s\n", get_real_address(0x204));
    break;
  case UADE_CHECKERROR:
    fprintf(stderr,"uade: FAILURE: Module check FAILED\n");
    if(!uade_slavemode)
      uade_reboot = 1;
    break;
  case UADE_SCORECRASH:
    fprintf(stderr,"uade: FAILURE: Soundcore CRASHED\n");
    if(uade_slavemode)
      uade_struct->songendboolean = 1;
    else
      uade_reboot = 1;
    break;
  case UADE_SCOREDEAD:
    fprintf(stderr,"uade: FAILURE: Soundcore is in DEADEND\n");
    if(uade_slavemode) {
      uade_struct->songendboolean = 1;
    } else {
      uade_reboot = 1;
    }
    break;
  case UADE_LOADFILE:
    /* load a file named at 0x204 (name pointer) to address pointed by
       0x208 and insert the length to 0x20C */
    src = uade_get_long(0x204);
    if(!valid_address(src, 1)) {
      fprintf(stderr, "uade: load: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    if((file = uade_open_amiga_file(nameptr))) {
      if ((status = decrunch (&file, nameptr)) < 0) {
	fprintf (stderr, "uade: load: decrunching error\n");
	fclose(file);
	break;
      } else {
	dst = uade_get_long(0x208);
	len = uade_safe_load(dst, file, uade_highmem - dst);
	fclose(file); file = 0;
	uade_put_long(0x20C, len);
	/* fprintf(stderr, "uade: load: %s: ptr = 0x%x size = 0x%x\n", nameptr, dst, len); */
      }
    }
    break;
  case UADE_READ:
    src = uade_get_long(0x204);
    if(!valid_address(src, 1)) {
      fprintf(stderr, "uade: read: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    dst = uade_get_long(0x208);
    off = uade_get_long(0x20C);
    len = uade_get_long(0x210);
    /* fprintf(stderr,"uade: read: '%s' dst = 0x%x off = 0x%x len = 0x%x\n", nameptr, dst, off, len); */
    if((file = uade_open_amiga_file(nameptr))) {
      fseek(file, off, SEEK_SET);
      x = uade_safe_load(dst, file, len);
      fclose(file); file = 0;
      if(!x)
	fprintf(stderr, "uade: read: 0x%x bytes read\n", x);
      uade_put_long(0x214, x);
    } else {
      fprintf(stderr, "uade: read: error when reading '%s'\n", nameptr);
      uade_put_long(0x214, 0);
    }
    break;
  case UADE_FILESIZE:
    src = uade_get_long(0x204);
    if(!valid_address(src, 1)) {
      fprintf(stderr, "uade: filesize: name in invalid address range\n");
      break;
    }
    nameptr = get_real_address(src);
    if((file = uade_open_amiga_file(nameptr))) {
      fseek(file, 0, SEEK_END);
      len = ftell(file);
      fclose(file); file = 0;
      /* fprintf(stderr, "uade: size: 0x%x '%s'\n", len, nameptr); */
      uade_put_long(0x208, len);
      uade_put_long(0x20C, -1);
    } else {
      fprintf(stderr, "uade: can't get file size for '%s'\n", nameptr);
      uade_put_long(0x208, 0);
      uade_put_long(0x20C, 0);
    }
    break;
  default:
    fprintf(stderr,"uade: UNKNOWN MESSAGE from score: $%x\n",x);
    break;
  }
}


static void uade_set_subsong(int subs) {
  uade_put_long(SCORE_SUBSONG, subs);
  uade_send_amiga_message(UADE_SETSUBSONG);
  uade_clear_soundbuffer();
}

static void uade_set_ntsc(int usentsc) {
  uade_put_long(SCORE_NTSC, usentsc);
}

static void uade_set_songend(int songendpossible) {
  uade_put_long(SCORE_HAVE_SONGEND, songendpossible);
}

static void uade_send_amiga_message(int msgtype) {
  uade_put_long(SCORE_OUTPUT_MSG, msgtype);
}

void uade_clear_soundbuffer(void) {
  if(uade_slavemode) {
    memset(uade_struct->soundbuffer, 0, sizeof(uade_struct->soundbuffer));
  }
}


static int uade_calc_reloc_size(unsigned int *src) {
  unsigned int offset;
  int i, nhunks;
  if(sizeof(unsigned int) != 4) {
    fprintf(stderr, "uade: reloc size problem: unsigned int != 4\n");
    fprintf(stderr, "uade: make reloc size calculator more portable\n");
    return 0;
  }
  if(ntohl(*src) != 0x03f3) return 0;
  src++;
  if(ntohl(*src)) return 0;
  src++;
  nhunks = ntohl(*src); src++; /* take number of hunks */
  src += 2;                    /* skip first & last hunk indexes */
  offset = 0;
  for(i = 0; i < nhunks; i++) {
    offset += 4 * (ntohl(*src) & 0x3FFFFFFF);
    src++;
  }
  return ((int) offset);
}
